@@ -125,12 +125,7 @@ struct impl {
 };
 
 struct tunnel_info {
-	AvahiIfIndex interface;
-	AvahiProtocol protocol;
 	const char *name;
-	const char *host_name;
-	const char *type;
-	const char *domain;
 };
 
 #define TUNNEL_INFO(...) ((struct tunnel_info){ __VA_ARGS__ })
@@ -152,12 +147,7 @@ static struct tunnel *make_tunnel(struct impl *impl, const struct tunnel_info *i
 	if (t == NULL)
 		return NULL;
 
-	t->info.interface = info->interface;
-	t->info.protocol = info->protocol;
 	t->info.name = strdup(info->name);
-	t->info.host_name = strdup(info->host_name);
-	t->info.type = strdup(info->type);
-	t->info.domain = strdup(info->domain);
 	spa_list_append(&impl->tunnel_list, &t->link);
 
 	return t;
@@ -167,11 +157,7 @@ static struct tunnel *find_tunnel(struct impl *impl, const struct tunnel_info *i
 {
 	struct tunnel *t;
 	spa_list_for_each(t, &impl->tunnel_list, link) {
-		if (t->info.interface == info->interface &&
-		    t->info.protocol == info->protocol &&
-		    spa_streq(t->info.name, info->name) &&
-		    spa_streq(t->info.type, info->type) &&
-		    spa_streq(t->info.domain, info->domain))
+		if (spa_streq(t->info.name, info->name))
 			return t;
 	}
 	return NULL;
@@ -179,7 +165,11 @@ static struct tunnel *find_tunnel(struct impl *impl, const struct tunnel_info *i
 
 static void free_tunnel(struct tunnel *t)
 {
-	pw_impl_module_destroy(t->module);
+	spa_list_remove(&t->link);
+	if (t->module)
+		pw_impl_module_destroy(t->module);
+	free((char *) t->info.name);
+	free(t);
 }
 
 static void impl_free(struct impl *impl)
@@ -292,16 +282,8 @@ static void pw_properties_from_avahi_string(const char *key, const char *value,
 static void submodule_destroy(void *data)
 {
 	struct tunnel *t = data;
-
-	spa_list_remove(&t->link);
 	spa_hook_remove(&t->module_listener);
-
-	free((char *) t->info.name);
-	free((char *) t->info.host_name);
-	free((char *) t->info.type);
-	free((char *) t->info.domain);
-
-	free(t);
+	t->module = NULL;
 }
 
 static const struct pw_impl_module_events submodule_events = {
@@ -312,19 +294,18 @@ static const struct pw_impl_module_events submodule_events = {
 struct match_info {
 	struct impl *impl;
 	struct pw_properties *props;
-	struct tunnel_info *tinfo;
+	struct tunnel *tunnel;
 	bool matched;
 };
 
 static int create_stream(struct impl *impl, struct pw_properties *props,
-		struct tunnel_info *tinfo)
+		struct tunnel *t)
 {
 	FILE *f;
 	char *args;
 	size_t size;
 	int res = 0;
 	struct pw_impl_module *mod;
-	struct tunnel *t;
 
 	if ((f = open_memstream(&args, &size)) == NULL) {
 		res = -errno;
@@ -349,16 +330,7 @@ static int create_stream(struct impl *impl, struct pw_properties *props,
                 goto done;
 	}
 
-	t = make_tunnel(impl, tinfo);
-	if (t == NULL) {
-		res = -errno;
-		pw_log_error("Can't make tunnel: %m");
-		pw_impl_module_destroy(mod);
-		goto done;
-	}
-
 	pw_impl_module_add_listener(mod, &t->module_listener, &submodule_events, t);
-
 	t->module = mod;
 done:
 	return res;
@@ -373,7 +345,7 @@ static int rule_matched(void *data, const char *location, const char *action,
 	i->matched = true;
 	if (spa_streq(action, "create-stream")) {
 		pw_properties_update_string(i->props, str, len);
-		create_stream(i->impl, i->props, i->tinfo);
+		create_stream(i->impl, i->props, i->tunnel);
 	}
 	return res;
 }
@@ -385,6 +357,7 @@ static void resolver_cb(AvahiServiceResolver *r, AvahiIfIndex interface, AvahiPr
 {
 	struct impl *impl = userdata;
 	struct tunnel_info tinfo;
+	struct tunnel *t;
 	const char *str;
 	AvahiStringList *l;
 	struct pw_properties *props = NULL;
@@ -396,12 +369,20 @@ static void resolver_cb(AvahiServiceResolver *r, AvahiIfIndex interface, AvahiPr
 				avahi_strerror(avahi_client_errno(impl->client)));
 		goto done;
 	}
-	tinfo = TUNNEL_INFO(.interface = interface,
-			.protocol = protocol,
-			.host_name = host_name,
-			.name = name,
-			.type = type,
-			.domain = domain);
+
+	tinfo = TUNNEL_INFO(.name = name);
+
+	t = find_tunnel(impl, &tinfo);
+	if (t == NULL)
+		t = make_tunnel(impl, &tinfo);
+	if (t == NULL) {
+		pw_log_error("Can't make tunnel: %m");
+		goto done;
+	}
+	if (t->module != NULL) {
+		pw_log_info("found duplicate mdns entry - skipping tunnel creation");
+		goto done;
+	}
 
 	props = pw_properties_new(NULL, NULL);
 	if (props == NULL) {
@@ -411,7 +392,6 @@ static void resolver_cb(AvahiServiceResolver *r, AvahiIfIndex interface, AvahiPr
 
 	avahi_address_snprint(at, sizeof(at), a);
 	ipv = protocol == AVAHI_PROTO_INET ? 4 : 6;
-
 	pw_properties_setf(props, "raop.ip", "%s", at);
 	pw_properties_setf(props, "raop.ip.version", "%d", ipv);
 	pw_properties_setf(props, "raop.port", "%u", port);
@@ -436,7 +416,7 @@ static void resolver_cb(AvahiServiceResolver *r, AvahiIfIndex interface, AvahiPr
 		struct match_info minfo = {
 			.impl = impl,
 			.props = props,
-			.tinfo = &tinfo,
+			.tunnel = t,
 		};
 		pw_conf_match_rules(str, strlen(str), NAME, &props->dict,
 				rule_matched, &minfo);
@@ -462,18 +442,16 @@ static void browser_cb(AvahiServiceBrowser *b, AvahiIfIndex interface, AvahiProt
 	if (flags & AVAHI_LOOKUP_RESULT_LOCAL)
 		return;
 
-	info = TUNNEL_INFO(.interface = interface,
-			.protocol = protocol,
-			.name = name,
-			.type = type,
-			.domain = domain);
+	info = TUNNEL_INFO(.name = name);
 
 	t = find_tunnel(impl, &info);
 
 	switch (event) {
 	case AVAHI_BROWSER_NEW:
-		if (t != NULL)
+		if (t != NULL) {
+			pw_log_info("found duplicate mdns entry - skipping tunnel creation");
 			return;
+		}
 		if (!(avahi_service_resolver_new(impl->client,
 						interface, protocol,
 						name, type, domain,
